using System;
using System.Collections.Generic;
using System.Text;

namespace TestSearch
{
    /// <summary>
    /// Indexes keys with sub-strings to allow fast searching of partial matches
    /// </summary>
    /// <typeparam name="T">Type of object being indexed</typeparam>
    class SubStringIndexer<T> : IIndexer<T>
    {
        private class IndexStruct
        {
            public string key;
            public T val;

            public IndexStruct(string key, T val)
            {
                this.key = key;
                this.val = val;
            }
        };

        #region Fields
        private int _maxKeyLength = 999;
        private Dictionary<int, List<IndexStruct>> _hashes;
        #endregion

        /// <summary>
        /// Initializes a new instance of the <see cref="SubStringIndexer&lt;T&gt;"/> class.
        /// </summary>
        /// <param name="numItems">The num items.</param>
        public SubStringIndexer(int numItems)
            : this(numItems, 999)
        {

        }

        public void FinishIndex()
        {
            //no need to do anything
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SubStringIndexer&lt;T&gt;"/> class.
        /// </summary>
        /// <param name="numItems">The number of items, initially.</param>
        /// <param name="maxKeyLength">Length of the max key.</param>
        public SubStringIndexer(int numItems, int maxKeyLength)
        {
            _maxKeyLength = maxKeyLength;
            _hashes = new Dictionary<int, List<IndexStruct>>(numItems);
        }
        

        #region IIndexer<T> Members

        /// <summary>
        /// Adds an item to the index
        /// </summary>
        /// <param name="key">The key.</param>
        /// <param name="value">The value.</param>
        public void AddItem(string key, T value, UInt32 sortOrder)
        {
            string toAdd = key.ToLower();

            toAdd = RemoveUnneededCharacters(toAdd);

            List<string> subStrings = GetSubStrings(toAdd);
            IndexStruct indexStruct = new IndexStruct(toAdd, value);
            foreach (string str in subStrings)
            {
                List<IndexStruct> items = null;
                int hash = str.GetHashCode();

                bool alreadyExists = _hashes.TryGetValue(hash, out items);
                if (!alreadyExists)
                {
                    items = new List<IndexStruct>();
                    _hashes[hash] = items;
                }
                items.Add(indexStruct);
            }
        }

        #region Key helpers
        /// <summary>
        /// Removes all but alphanumeric characters from the input string
        /// </summary>
        /// <param name="original">The original.</param>
        /// <returns>A string with all unneeded characters removed</returns>
        private string RemoveUnneededCharacters(string original)
        {
            char[] array = new char[original.Length];
            int destIndex = 0;
            for (int i = 0; i < original.Length; i++)
            {
                char c = original[i];
                if (char.IsLetterOrDigit(c))
                {
                    array[destIndex] = c;
                    destIndex++;
                }
            }
            return new string(array, 0, destIndex);
        }


        /// <summary>
        /// Gets a list of all indexable substrings.
        /// </summary>
        /// <param name="key">The key.</param>
        /// <returns>List of substrings</returns>
        private List<string> GetSubStrings(string key)
        {
            List<string> results = new List<string>();

            for (int start = 0; start < key.Length; start++)
            {
                /*get maximum length of substring based on current
                 * character position (constrain it to within the string
                 * and less than or equal to the maximum key length specified
                 * */
                int lastLength = Math.Min(key.Length - start, _maxKeyLength);

                /* Get each substring from length 1 to lastLength
                 */
                for (int length = 1; length <= lastLength; length++)
                {
                    string sub = key.Substring(start, length);
                    if (!results.Contains(sub))
                    {
                        results.Add(sub);
                    }
                }
            }
            return results;
        }
        #endregion

        /// <summary>
        /// Looks up the specified sub key.
        /// </summary>
        /// <param name="subKey">The sub key.</param>
        /// <returns>A list of values that match the subkey</returns>
        public IList<T> Lookup(string subKey)
        {
            string toLookup = subKey.ToLower();
            List<IndexStruct> items = null;
            List<T> results = new List<T>();
            int hash = 0;

            if (subKey.Length > _maxKeyLength)
            {
                /*
                 * If the substring is too long, get the longest substring 
                 * we've indexed and use that for the initial search
                 */
                toLookup = toLookup.Substring(0, _maxKeyLength);
                hash = toLookup.GetHashCode();
            }
            else
            {
                hash = toLookup.GetHashCode();
            }

            bool found = _hashes.TryGetValue(hash, out items);
            if (found)
            {
                results.Capacity = items.Count;
                foreach (IndexStruct s in items)
                {
                    /*
                     * Have to check each item in this bucket's list
                     * because the substring might be longer than the indexed
                     * keys
                     */
                    if (s.key.IndexOf(subKey, StringComparison.InvariantCultureIgnoreCase) >= 0)
                    {
                        results.Add(s.val);
                    }
                }
            }
            
            return results;
        }


        #endregion


#if NUNIT
        internal List<string> NUnit_GetSubStrings(string key)
        {
            return GetSubStrings(key);
        }
#endif
    }
}
